import os
import sys
import pickle

import numpy as np
import wx
import wx.lib.filebrowsebutton as filebrowse
from wx.lib.expando import ExpandoTextCtrl, EVT_ETC_LAYOUT_NEEDED

from .. import art
from ..utils.processutil import which
from ..utils.textutil import text_to_int
from .dropscroller import ReorderableList, PickledDropTarget, PickledDataObject

import logging
log = logging.getLogger(__name__)


class SimplePromptDialog(wx.TextEntryDialog):
    """Simple subclass of wx.TextEntryDialog to convert text result to a
    specific format
    """

    def convert_text(self, text, return_error=False, **kwargs):
        if return_error:
            return text, ""
        else:
            return text

    def show_and_get_value(self, return_error=False, **kwargs):
        result = self.ShowModal()
        if result == wx.ID_OK:
            text = self.GetValue()
            value = self.convert_text(text, return_error, **kwargs)
        else:
            if return_error:
                value = None, "Cancelled"
            else:
                value = None
        self.Destroy()
        return value


class HexEntryDialog(SimplePromptDialog):
    """Simple subclass of wx.TextEntryDialog to convert text result from
    hexidecimal if necessary.
    """

    def convert_text(self, text, return_error=False, default_base="dec", **kwargs):
        try:
            value = text_to_int(text, default_base)
            error = ""
        except (ValueError, TypeError) as e:
            value = None
            error = str(e)
        if return_error:
            return value, error
        else:
            return value


class SliceEntryDialog(SimplePromptDialog):
    """Simple subclass of wx.TextEntryDialog to convert text result from
    hexidecimal if necessary.
    """

    def convert_text(self, text, return_error=False, default_base="dec", **kwargs):
        try:
            if "," in text:
                start, extra = text.split(",")
                if "," in extra:
                    end, step = extra.split(",")
                else:
                    end = extra
                    step = "1"
            else:
                start = text
                end = "-1"
                step = "1"
            start = text_to_int(start, default_base)
            end = text_to_int(end, default_base)
            step = text_to_int(step, default_base)
            error = ""
        except (ValueError, TypeError) as e:
            start, end, step = None, None, None
            error = str(e)
        if return_error:
            return None, error
        else:
            return slice(start, end, step)


def prompt_for_string(parent, message, title, default=None):
    if default is not None:
        default = str(default)
    else:
        default = ""
    d = SimplePromptDialog(parent, message, title, default)
    return d.show_and_get_value()


def prompt_for_hex(parent, message, title, default=None, return_error=False, default_base="hex"):
    if default is not None:
        if default_base == "hex":
            default = hex(int(default))[2:]
        else:
            default = str(int(default))
    else:
        default = ""
    d = HexEntryDialog(parent, message, title, default)
    if default:
        d.SetValue(default)
    return d.show_and_get_value(return_error=return_error, default_base=default_base)


def prompt_for_dec(parent, message, title, default=None, return_error=False, default_base="dec"):
    return prompt_for_hex(parent, message, title, default, return_error, default_base)


def prompt_for_slice(parent, message, title, default=None, return_error=False, default_base="hex"):
    if default is not None:
        if default_base == "hex":
            default = hex(int(default))[2:]
        else:
            default = str(int(default))
    else:
        default = ""
    d = SliceEntryDialog(parent, message, title, default)
    if default:
        d.SetValue(default)
    return d.show_and_get_value(return_error=return_error, default_base=default_base)


class DictEditDialog(wx.Dialog):
    border = 5

    def __init__(self, parent, title, instructions, fields, default=None):
        wx.Dialog.__init__(self, parent, -1, title)
        sizer = wx.BoxSizer(wx.VERTICAL)
        self.SetSizer(sizer)

        t = wx.StaticText(self, -1, instructions)
        sizer.Add(t, 0, wx.ALL|wx.EXPAND, self.border)

        self.types = {}
        self.controls = {}
        self.buttons = {}
        self.add_fields(fields)

        btnsizer = wx.StdDialogButtonSizer()
        self.ok_btn = wx.Button(self, wx.ID_OK)
        self.ok_btn.SetDefault()
        btnsizer.AddButton(self.ok_btn)
        btn = wx.Button(self, wx.ID_CANCEL)
        btnsizer.AddButton(btn)
        btnsizer.Realize()
        sizer.Add(btnsizer, 1, wx.ALL|wx.EXPAND, self.border)

        self.Bind(wx.EVT_BUTTON, self.on_button)
        self.Bind(wx.EVT_TEXT, self.on_text_changed)
        self.Bind(EVT_ETC_LAYOUT_NEEDED, self.on_resize)

        # Don't call self.Fit() otherwise the dialog buttons are zero height
        sizer.Fit(self)

        self.default = default
        if default:
            self.set_initial_values(default)
        self.check_enable()

    def add_fields(self, fields):
        """ Calls the create_X method where X is the type of the field. Fields
        are a list of tuples: (type, key, label) or (type, key, label, choices)
        """
        for field in fields:
            try:
                type, key, label = field
                choices = None
            except ValueError:
                type, key, label, choices = field
            try:
                func = getattr(self, "create_%s" % type.replace(" ", "_"))
            except AttributeError:
                raise NotImplementedError("Unknown field type %s for %s" % (type, key))
            control = func(key, label, choices)
            if key is not None:
                self.types[key] = type
                self.controls[key] = control

    def set_initial_values(self, d):
        for key in list(self.controls.keys()):
            self.set_initial_value_of(d, key)

    def set_initial_value_of(self, d, key):
        control = self.controls[key]
        type = self.types[key]
        if type == 'text' or type == 'verify':
            value = self.get_default_value(d, key)
            control.ChangeValue(value)
        elif type == 'verify list':
            value = "\n".join(self.get_default_value(d, key))
            control.ChangeValue(value)
        elif type == 'file' or type == 'boolean':
            value = self.get_default_value(d, key)
            control.SetValue(value)
        elif type == 'dropdown':
            value = self.get_default_value(d, key)
            control.SetStringSelection(value)

    def get_default_value(self, d, key):
        return d[key]

    def get_edited_values(self, d):
        for key in list(self.controls.keys()):
            self.get_edited_value_of(d, key)

    def get_edited_value_of(self, d, key):
        control = self.controls[key]
        type = self.types[key]
        try:
            if type == 'verify list':
                value = control.GetValue().splitlines()
            elif type == 'dropdown':
                value = control.GetStringSelection()
            else:
                value = control.GetValue()
            self.set_output_value(d, key, value)
        except AttributeError:
            log.error("Error setting output value for %s %s" % (key, control))

    def set_output_value(self, d, key, value):
        d[key] = value

    def create_text(self, key, label, choices=None):
        sizer = self.GetSizer()
        hbox = wx.BoxSizer(wx.HORIZONTAL)
        t = wx.StaticText(self, -1, label)
        hbox.Add(t, 0, wx.LEFT|wx.ALIGN_CENTER_VERTICAL, self.border)
        entry = wx.TextCtrl(self, -1, size=(-1, -1))
        hbox.Add(entry, 1, wx.ALL|wx.EXPAND, self.border)
        sizer.Add(hbox, 0, wx.LEFT|wx.RIGHT|wx.TOP|wx.EXPAND, self.border)
        return entry

    def create_boolean(self, key, label, choices=None):
        sizer = self.GetSizer()
        hbox = wx.BoxSizer(wx.HORIZONTAL)
        t = wx.StaticText(self, -1, label)
        hbox.Add(t, 0, wx.LEFT|wx.ALIGN_CENTER_VERTICAL, self.border)
        entry = wx.CheckBox(self, -1, size=(-1, -1))
        hbox.Add(entry, 1, wx.ALL|wx.EXPAND, self.border)
        sizer.Add(hbox, 0, wx.LEFT|wx.RIGHT|wx.TOP|wx.EXPAND, self.border)
        return entry

    def create_expando(self, key, label, choices=None):
        sizer = self.GetSizer()
        status = ExpandoTextCtrl(self, style=wx.ALIGN_LEFT|wx.TE_READONLY|wx.NO_BORDER)
        attr = self.GetDefaultAttributes()
        status.SetBackgroundColour(attr.colBg)
        sizer.Add(status, 1, wx.ALL|wx.EXPAND, self.border)
        return status

    def create_file(self, key, label, choices=None):
        sizer = self.GetSizer()
        hbox = wx.BoxSizer(wx.HORIZONTAL)
        entry = filebrowse.FileBrowseButton(self, -1, size=(450, -1), labelText=label, changeCallback=self.on_path_changed)
        hbox.Add(entry, 0, wx.LEFT, self.border)
        sizer.Add(hbox, 0, wx.ALL|wx.EXPAND, 0)
        return entry

    def create_verify(self, key, label, choices=None):
        sizer = self.GetSizer()
        hbox = wx.BoxSizer(wx.HORIZONTAL)
        t = wx.StaticText(self, -1, label)
        hbox.Add(t, 0, wx.LEFT|wx.ALIGN_CENTER_VERTICAL, self.border)
        entry = wx.TextCtrl(self, -1, size=(-1, -1))
        hbox.Add(entry, 1, wx.ALL|wx.EXPAND, self.border)
        verify = wx.Button(self, -1, "Verify")
        hbox.Add(verify, 0, wx.LEFT|wx.RIGHT|wx.TOP|wx.EXPAND|wx.ALIGN_CENTER, self.border)
        sizer.Add(hbox, 0, wx.LEFT|wx.RIGHT|wx.TOP|wx.EXPAND, self.border)
        verify.Bind(wx.EVT_BUTTON, self.on_verify)
        self.buttons[key] = verify
        return entry

    def create_verify_list(self, key, label, choices=None):
        sizer = self.GetSizer()
        hbox = wx.BoxSizer(wx.HORIZONTAL)
        t = wx.StaticText(self, -1, label)
        hbox.Add(t, 0, wx.LEFT|wx.ALIGN_CENTER_VERTICAL, self.border)
        entry = wx.TextCtrl(self, -1, size=(-1, 100), style=wx.TE_MULTILINE|wx.HSCROLL)
        hbox.Add(entry, 1, wx.ALL|wx.EXPAND, self.border)
        verify = wx.Button(self, -1, "Verify")
        hbox.Add(verify, 0, wx.LEFT|wx.RIGHT|wx.ALIGN_CENTER, self.border)
        sizer.Add(hbox, 0, wx.LEFT|wx.RIGHT|wx.TOP|wx.EXPAND, self.border)
        verify.Bind(wx.EVT_BUTTON, self.on_verify)
        self.buttons[key] = verify
        return entry

    def create_gauge(self, key, label, choices=None):
        sizer = self.GetSizer()
        gauge = wx.Gauge(self, -1, 50, size=(500, 5))
        sizer.Add(gauge, 0, wx.ALL|wx.EXPAND, self.border)
        return gauge

    def create_static(self, key, label, choices=None):
        sizer = self.GetSizer()
        t = wx.StaticText(self, -1, label)
        sizer.Add(t, 0, wx.LEFT|wx.RIGHT|wx.BOTTOM|wx.EXPAND, self.border)

    def create_static_spacer_below(self, key, label, choices=None):
        sizer = self.GetSizer()
        t = wx.StaticText(self, -1, label)
        sizer.Add(t, 0, wx.LEFT|wx.RIGHT|wx.BOTTOM|wx.EXPAND, self.border)
        t = wx.StaticText(self, -1, " ")
        sizer.Add(t, 0, wx.ALL|wx.EXPAND, self.border)

    def create_dropdown(self, key, label, choices):
        sizer = self.GetSizer()
        hbox = wx.BoxSizer(wx.HORIZONTAL)
        t = wx.StaticText(self, -1, label)
        hbox.Add(t, 0, wx.LEFT|wx.ALIGN_CENTER_VERTICAL, self.border)
        entry = wx.Choice(self, -1, choices=choices, size=(-1, -1))
        hbox.Add(entry, 1, wx.ALL|wx.EXPAND, self.border)
        sizer.Add(hbox, 0, wx.LEFT|wx.RIGHT|wx.TOP|wx.EXPAND, self.border)
        return entry

    def on_button(self, evt):
        if evt.GetId() == wx.ID_OK:
            self.EndModal(wx.ID_OK)
        else:
            self.EndModal(wx.ID_CANCEL)
        evt.Skip()

    def on_text_changed(self, evt):
        self.ok_btn.Enable(self.can_submit())

    def on_path_changed(self, evt):
        pass

    def on_verify(self, evt):
        pass

    def on_resize(self, event):
        self.Fit()

    def can_submit(self):
        return True

    def check_enable(self):
        self.ok_btn.Enable(self.can_submit())

    def show_and_get_value(self):
        result = self.ShowModal()
        if result == wx.ID_OK:
            d = self.get_result_object()
            self.get_edited_values(d)
        else:
            d = None
        self.Destroy()
        return d

    def get_result_object(self):
        # Edit the object in place by reusing the same dictionary
        if self.default:
            d = self.default
        else:
            d = self.get_new_object()
        return d

    def get_new_object(self):
        return dict()


class ObjectEditDialog(DictEditDialog):
    def __init__(self, parent, title, instructions, fields, object_class, default=None):
        DictEditDialog.__init__(self, parent, title, instructions, fields, default)
        self.new_object_class = object_class

    def get_default_value(self, d, key):
        return getattr(d, key)

    def get_new_object(self):
        return self.new_object_class()

    def set_output_value(self, d, key, value):
        setattr(d, key, value)


def get_file_dialog_wildcard(extension_list):
    wildcards = []
    for item in extension_list:
        name = item[0]
        exts = item[1:]
        ext = ";".join(["*" + e for e in exts])
        wildcards.append("%s (%s)|%s" % (name, ext, ext))
    return "|".join(wildcards)


class ListReorderDialog(wx.Dialog):
    """Simple dialog to return a list of items that can be reordered by the user.
    """
    border = 5

    def __init__(self, parent, items, get_item_text, dialog_helper=None, title="Reorder List", copy_helper=None, default_helper=None):
        wx.Dialog.__init__(self, parent, -1, title,
                           size=(700, 500), pos=wx.DefaultPosition,
                           style=wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER)

        sizer = wx.BoxSizer(wx.HORIZONTAL)

        self.list = ReorderableList(self, items, get_item_text, size=(-1,500))
        self.list.Bind(wx.EVT_LIST_ITEM_SELECTED, self.on_list_selection)
        self.list.Bind(wx.EVT_LIST_ITEM_DESELECTED, self.on_list_selection)
        sizer.Add(self.list, 1, wx.EXPAND)

        vbox = wx.BoxSizer(wx.VERTICAL)

        self.up = b = wx.Button(self, wx.ID_UP, 'Up', size=(90, -1))
        b.Bind(wx.EVT_BUTTON, self.on_up)
        vbox.Add(b, 0, wx.ALL|wx.EXPAND, self.border)

        self.down = b = wx.Button(self, wx.ID_DOWN, 'Down', size=(90, -1))
        b.Bind(wx.EVT_BUTTON, self.on_down)
        vbox.Add(b, 0, wx.ALL|wx.EXPAND, self.border)

        vbox.AddSpacer(50)

        b = wx.Button(self, wx.ID_NEW, 'New', size=(90, -1))
        b.Bind(wx.EVT_BUTTON, self.on_new)
        vbox.Add(b, 0, wx.ALL|wx.EXPAND, self.border)

        self.edit = b = wx.Button(self, wx.ID_EDIT, 'Edit', size=(90, -1))
        b.Bind(wx.EVT_BUTTON, self.on_edit)
        vbox.Add(b, 0, wx.ALL|wx.EXPAND, self.border)

        if copy_helper is not None:
            self.copy = b = wx.Button(self, wx.ID_COPY, 'Copy', size=(90, -1))
            b.Bind(wx.EVT_BUTTON, self.on_copy)
            vbox.Add(b, 0, wx.ALL|wx.EXPAND, self.border)
        else:
            self.copy = None

        if default_helper is not None:
            self.default = b = wx.Button(self, wx.ID_COPY, 'Set as Default', size=(90, -1))
            b.Bind(wx.EVT_BUTTON, self.on_set_default)
            vbox.Add(b, 0, wx.ALL|wx.EXPAND, self.border)
        else:
            self.default = None
        self.default_helper = default_helper

        vbox.AddSpacer(50)

        self.delete = b = wx.Button(self, wx.ID_DELETE, 'Delete', size=(90, -1))
        b.Bind(wx.EVT_BUTTON, self.on_delete)
        vbox.Add(b, 0, wx.ALL|wx.EXPAND, self.border)

        vbox.AddStretchSpacer()

        btnsizer = wx.StdDialogButtonSizer()
        btn = wx.Button(self, wx.ID_OK)
        btn.SetDefault()
        btnsizer.AddButton(btn)
        btn = wx.Button(self, wx.ID_CANCEL)
        btnsizer.AddButton(btn)
        btnsizer.Realize()
        vbox.Add(btnsizer, 0, wx.ALL|wx.EXPAND, self.border)
        sizer.Add(vbox, 0, wx.EXPAND, 0)

        self.SetSizer(sizer)
        sizer.Fit(self)

        self.Layout()

        self.list.Bind(wx.EVT_CONTEXT_MENU, self.on_context_menu)
        self.delete_id = wx.NewId()
        self.Bind(wx.EVT_MENU, self.on_delete, id=self.delete_id)

        self.get_item_text = get_item_text
        self.dialog_helper = dialog_helper
        self.copy_helper = copy_helper
        self.on_list_selection(None)

    def on_list_selection(self, evt):
        one_selected = self.list.GetSelectedItemCount() == 1
        any_selected = self.list.GetSelectedItemCount() > 0
        self.up.Enable(self.list.can_move_up)
        self.down.Enable(self.list.can_move_down)
        self.edit.Enable(one_selected)
        if self.copy is not None:
            self.copy.Enable(one_selected)
        if self.default is not None:
            self.default.Enable(one_selected)
        self.delete.Enable(any_selected)

    def on_context_menu(self, evt):
        one_selected = self.list.GetSelectedItemCount() == 1
        any_selected = self.list.GetSelectedItemCount() > 0
        menu = wx.Menu()
        menu.Append(wx.ID_NEW, "New Item")
        menu.Append(wx.ID_EDIT, "Edit Item")
        menu.Enable(wx.ID_EDIT, one_selected)
        if self.copy is not None:
            menu.Append(wx.ID_COPY, "Copy Item")
            menu.Enable(wx.ID_COPY, one_selected)
        menu.AppendSeparator()
        menu.Append(wx.ID_SELECTALL, "Select All")
        menu.Append(wx.ID_CLEAR, "Deselect All")
        menu.Append(wx.ID_DELETE, "Delete Selected Items")
        menu.Enable(wx.ID_DELETE, any_selected)
        id = self.GetPopupMenuSelectionFromUser(menu)
        menu.Destroy()
        if id == wx.ID_NEW:
            self.on_new(evt)
        elif id == wx.ID_DELETE:
            self.on_delete(evt)
        elif id == wx.ID_EDIT:
            self.on_edit(evt)
        elif id == wx.ID_COPY:
            self.on_copy(evt)
        if id == wx.ID_SELECTALL:
            self.list.select_all()
        elif id == wx.ID_CLEAR:
            self.list.deselect_all()

    def get_items(self):
        return self.list.items

    def on_up(self, evt):
        if self.list.can_move_up:
            self.list.move_selected(-1)

    def on_down(self, evt):
        if self.list.can_move_down:
            self.list.move_selected(1)

    def on_new(self, evt):
        new_item = self.dialog_helper(self, "Add Item")
        if new_item is not None:
            self.insert_new_item(new_item)

    def insert_new_item(self, new_item):
        index = self.list.GetFirstSelected()
        if index == -1:
            index = len(self.list.items)
        else:
            index += 1
        self.list.items[index:index] = [new_item]
        self.list.refresh()
        self.list.deselect_all()
        self.list.SetItemState(index, wx.LIST_STATE_SELECTED, wx.LIST_STATE_SELECTED)

    def on_edit(self, evt):
        index = self.list.GetFirstSelected()
        if index >= 0:
            item = self.list.items[index]
            new_item = self.dialog_helper(self, "Edit %s" % self.get_item_text(item), item)
            if new_item is not None:
                self.list.items[index] = new_item
                self.list.refresh()

    def on_copy(self, evt):
        index = self.list.GetFirstSelected()
        if index >= 0:
            item = self.list.items[index]
            new_item = self.copy_helper(item)
            if new_item is not None:
                self.insert_new_item(new_item)

    def on_set_default(self, evt):
        index = self.list.GetFirstSelected()
        if index >= 0:
            for i, item in enumerate(self.list.items):
                self.default_helper(item, i == index)
            self.list.refresh()

    def on_delete(self, evt):
        self.list.delete_selected()


class CheckItemDialog(wx.Dialog):
    """Simple dialog to return a list of items that can be reordered by the user.
    """
    border = 5

    def __init__(self, parent, items, get_item_text, dialog_helper=None, title="Select Items", instructions=""):
        wx.Dialog.__init__(self, parent, -1, title,
                           size=(700, 500), pos=wx.DefaultPosition,
                           style=wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER)

        sizer = wx.BoxSizer(wx.VERTICAL)

        self.get_item_text = get_item_text
        self.dialog_helper = dialog_helper

        if instructions:
            t = wx.StaticText(self, -1, instructions)
            sizer.Add(t, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, self.border)

        # Use multi-select to make sure no initial item is selecte. Single
        # select forces the first item to be selected.
        self.list = wx.CheckListBox(self, size=(-1,500), style=wx.LB_MULTIPLE)

        # Both EVT_LISTBOX and EVT_CHECKLISTBOX cancel each other out when
        # clicking on the check box itself. Clicking only on item text causes
        # EVT_LISTBOX event only.
#        self.list.Bind(wx.EVT_LISTBOX, self.on_list_selection)
        self.list.Bind(wx.EVT_CHECKLISTBOX, self.on_list_check)
        self.list.Bind(wx.EVT_CONTEXT_MENU, self.on_context_menu)
        sizer.Add(self.list, 1, wx.EXPAND)
        self.set_items(items)

        btnsizer = wx.StdDialogButtonSizer()
        btn = wx.Button(self, wx.ID_OK)
        btn.SetDefault()
        btnsizer.AddButton(btn)
        btn = wx.Button(self, wx.ID_CANCEL)
        btnsizer.AddButton(btn)
        btnsizer.Realize()
        sizer.Add(btnsizer, 0, wx.EXPAND, 0)

        self.SetSizer(sizer)
        sizer.Fit(self)

        self.Layout()

        self.Bind(wx.EVT_CONTEXT_MENU, self.on_context_menu)
        self.delete_id = wx.NewId()

    def set_items(self, items):
        self.clear()
        for item in items:
            self.insert_item(self.list.GetCount(), item)

    def insert_item(self, index, item):
        self.list.Insert("placeholder", index)
        self.set_item_text(index, item)
        self.items[index:index] = [item]

    def set_item_text(self, index, item):
        text = self.get_item_text(item)
        self.list.SetString(index, text)

    def on_list_selection(self, evt):
        index = evt.GetInt()
        self.list.Check(index, not self.list.IsChecked(index))
        self.list.SetSelection(index)

    def on_list_check(self, evt):
        index = evt.GetInt()

    def on_context_menu(self, evt):
        menu = wx.Menu()
        menu.Append(wx.ID_SELECTALL, "Select All")
        menu.Append(wx.ID_CLEAR, "Deselect All")
        id = self.GetPopupMenuSelectionFromUser(menu)
        menu.Destroy()
        if id == wx.ID_SELECTALL:
            self.select_all()
        elif id == wx.ID_CLEAR:
            self.deselect_all()

    def select_all(self):
        self.list.SetChecked(list(range(self.list.GetCount())))

    def deselect_all(self):
        self.list.SetChecked([])

    def get_items(self):
        return self.items

    def get_checked_items(self):
        return [self.items[i] for i in self.list.GetCheckedItems()]

    def clear(self, evt=None):
        self.list.Clear()
        self.items = []


class ChooseOnePlusCustomDialog(wx.Dialog):
    """Simple dialog to return a choice from a list or a custom value
    """
    border = 5

    def __init__(self, parent, items, default=None, default_custom_value="", title="Select Items", instructions="", custom_value_label="Custom Value"):
        wx.Dialog.__init__(self, parent, -1, title,
                           size=(700, 500), pos=wx.DefaultPosition,
                           style=wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER)

        sizer = wx.BoxSizer(wx.VERTICAL)

        if instructions:
            t = wx.StaticText(self, -1, instructions)
            sizer.Add(t, 0, wx.ALL|wx.ALIGN_CENTER_VERTICAL, self.border)

        # custom item will be last in the list
        self.custom_item_index = len(items)
        items.append("Custom")

        self.list = wx.ListBox(self, choices=items, size=(-1,200), style=wx.LB_SINGLE)
        try:
            index = items.index(default)
        except ValueError:
            if default_custom_value:
                index = self.custom_item_index
            else:
                index = 0
        self.list.SetSelection(index)

        # Both EVT_LISTBOX and EVT_CHECKLISTBOX cancel each other out when
        # clicking on the check box itself. Clicking only on item text causes
        # EVT_LISTBOX event only.
        self.list.Bind(wx.EVT_LISTBOX, self.on_list_selection)
        sizer.Add(self.list, 1, wx.EXPAND)

        hbox = wx.BoxSizer(wx.HORIZONTAL)
        t = wx.StaticText(self, -1, custom_value_label + ":")
        hbox.Add(t, 0, wx.LEFT|wx.ALIGN_CENTER_VERTICAL, self.border)
        self.custom_value = wx.TextCtrl(self, -1, default_custom_value, size=(-1, -1))
        hbox.Add(self.custom_value, 1, wx.ALL|wx.EXPAND, self.border)
        sizer.Add(hbox, 0, wx.EXPAND)

        btnsizer = wx.StdDialogButtonSizer()
        btn = wx.Button(self, wx.ID_OK)
        btn.SetDefault()
        btnsizer.AddButton(btn)
        btn = wx.Button(self, wx.ID_CANCEL)
        btnsizer.AddButton(btn)
        btnsizer.Realize()
        sizer.Add(btnsizer, 0, wx.EXPAND, 0)

        self.SetSizer(sizer)
        sizer.Fit(self)

        self.Layout()

    def on_list_selection(self, evt):
        index = evt.GetInt()
        if index == self.custom_item_index:
            wx.CallAfter(self.custom_value.SetFocus)
            self.custom_value.Enable(True)
        else:
            self.custom_value.Enable(False)
        evt.Skip()

    def get_selected(self):
        index = self.list.GetSelection()
        if index == self.custom_item_index:
            # custom value
            label = None
            custom = self.custom_value.GetValue()
        else:
            label = self.list.GetString(index)
            custom = None
        return label, custom


class HtmlAboutDialog(wx.Dialog):
    """Simple 'About' dialog with an image on the left and HTML text on
    the right.
    """
    border = 20

    def __init__(self, parent, title, html, bitmap):
        wx.Dialog.__init__(self, parent, -1, title,
                           size=(700, 500), pos=wx.DefaultPosition,
                           style=wx.DEFAULT_DIALOG_STYLE|wx.RESIZE_BORDER)
        self.SetBackgroundColour("#222255")
        sizer = wx.BoxSizer(wx.HORIZONTAL)

        self.bitmap = wx.StaticBitmap(self, -1, bitmap)
        sizer.Add(self.bitmap, 0, wx.ALIGN_TOP|wx.ALL, self.border)

        self.html = wx.html.HtmlWindow(self, -1, size=(400, 400))
        self.html.SetPage(html)
        sizer.Add(self.html, 1, wx.EXPAND|wx.ALL, 0)

        self.SetSizer(sizer)
        sizer.Fit(self)

        self.Layout()

        self.Bind(wx.EVT_ACTIVATE, self.on_activate)
        self.Bind(wx.html.EVT_HTML_LINK_CLICKED, self.on_link_clicked)

    def on_activate(self, evt):
        log.debug(f"on_link_clicked: {evt.GetActive()}")
        if not evt.GetActive():
            wx.CallAfter(self.Destroy)

    def on_link_clicked(self, evt):
        href = evt.GetLinkInfo().GetHref()
        log.debug(f"on_link_clicked: {href}")


class SawxAboutDialog(HtmlAboutDialog):
    @classmethod
    def is_shown(cls):
        for frame in wx.GetTopLevelWindows():
            if isinstance(frame, HtmlAboutDialog):
                return frame
        return None

    @classmethod
    def show_or_raise(cls):
        frame = cls.is_shown()
        if frame:
            frame.Raise()
        else:
            frame = cls()

    def __init__(self):
        app = wx.GetApp()
        bitmap = art.get_bitmap(app.about_dialog_image)
        html = app.about_dialog_html
        HtmlAboutDialog.__init__(self, None, app.app_name, html, bitmap)
        wx.CallAfter(self.Show)


if __name__ == "__main__":
    app = wx.PySimpleApp()

    frame = wx.Frame(None, -1, "Dialog test")
#    dialog = EmulatorDialog(frame, "Test")
#    dialog.ShowModal()
#    dlg = ListReorderDialog(frame, [chr(i + 65) for i in range(26)], lambda a:str(a))
    # dlg = CheckItemDialog(frame, [chr(i + 65) for i in range(26)], lambda a:str(a))
    # if dlg.ShowModal() == wx.ID_OK:
    #     print dlg.get_checked_items()
    # dlg = ChooseOnePlusCustomDialog(frame, ["one", "two"], "instructions")
    # if dlg.ShowModal() == wx.ID_OK:
    #     print "Selected", dlg.get_selected()
    # dlg.Destroy()

    class TestObj(object):
        def __init__(self):
            self.name = "aoeu"
            self._state = False

        @property
        def state(self):
            return 'state true' if self._state else 'state false'

        @state.setter
        def state(self, value):
            print(("Setting to:", value))
            self._state = (value == 'state true')

    class TestSetattrDialog(ObjectEditDialog):
        def __init__(self, parent, title, default=None):
            fields = [
                ('text', 'name', 'Server Name: '),
                ('dropdown', 'state', 'State choice', ['state true', 'state false']),
                ]
            ObjectEditDialog.__init__(self, parent, title, "Setattr test", fields, TestObj, default)

    test_obj = TestObj()
    dlg = TestSetattrDialog(frame, "Test", test_obj)
    dlg.show_and_get_value()
    print((test_obj.state))
    dlg = TestSetattrDialog(frame, "Test", test_obj)
    dlg.show_and_get_value()
    print((test_obj.state))

    app.MainLoop()



"""Popup window created as a child of a window and controlled by showing and
hiding the subwindow.

"""

import os, sys, struct, queue, threading, time, socket
from io import StringIO

import wx
import wx.stc

import logging
log = logging.getLogger(__name__)


class PopupSubWindow(wx.Window):
    """Transient status bar that displays status text in a popup
    
    Unlike a wx.StatusBar window that uses a constant amount of screen real-
    estate, the PopupStatusBar displays status info in a temporary popup
    that is displayed on the bottom of the frame.
    
    Status text is always overwritten with updates to the status text.  Status
    text is designed to be used for repetitive updates in response to an event;
    for example, to update coordinates when the mouse is moved.
    """
    def __init__(self, parent, target, side=wx.LEFT|wx.BOTTOM):
        """Creates (but doesn't show) the PopupStatusBar
        
        target is a sibling window; that is: target and this window share the
        same parent
        """
        wx.Window.__init__(self, parent, -1, name="PopupStatusWindow of %s" % parent.GetName(), pos=(0,0), size=(200, 15))
        self.target_window = target
        self.SetBackgroundColour("#B6C1FF")

        self.status = wx.StaticText(self, -1, "", style = wx.BORDER_NONE)
        self.Hide()

    def show_status_text(self, text, multiline=False):
        """Display a status text string in the status popup.
        
        This method is intended to display text in response to a large number
        of updates to a similar actions, for example: updating x,y coordinates
        in response to mouse movement.  It is undesirable to keep these
        messages in the list as the list would quickly grow to display many
        lines.  Instead, status text updates replace any previous status
        updates at the bottom of the popup.
        
        This forces the popup to be displayed if it isn't currently displayed.
        If the popup is displayed and other messages are present, the existing
        messages are moved up and this status text is inserted at the bottom.
        
        @param text: message to display
        """
        if multiline:
            text = text.replace("\r\n", "\n").replace("\n", "\n").replace("\r", "\n")
        else:
            text = text.replace("\r\n", " ").replace("\n", " ").replace("\r", " ")
        self.status.SetLabelText(text)
        if text:
            self.position_and_show()
        else:
            self.Hide()
            pass

    def position_and_show(self):
        target = self.target_window
        tx, ty = target.GetPosition()
        tw, th = target.GetSize()
        w, h = self.GetSize()
        x = tx
        y = ty + th - h
        self.SetPosition((x, y))
        if w > tw:
            self.SetSize(tw, h)
        log.debug("target pos:%s, size=%s  popup pos:%s size=%s  text=%s" % (str((tx, ty)), str((tw, th)), str((x, y)), str((w, h)), self.status.GetLabelText()))
        self.Show(True)


    def clear(self):
        """Remove all messages and hide the popup"""
        self.show_status_text("")


if __name__ == "__main__":
    import wx.stc

    logging.basicConfig(level=logging.DEBUG)

    class TestSTC(wx.stc.StyledTextCtrl):
        def __init__(self, *args, **kwargs):
            wx.stc.StyledTextCtrl.__init__(self, *args, **kwargs)
            self.Bind(wx.stc.EVT_STC_UPDATEUI, self.OnUpdateUI)

        def OnUpdateUI(self, evt):
            """Specific OnUpdateUI callback for those modes that use an actual
            STC for their edit window.
            
            Adds things like fold level and style display.
            """
            linenum = self.GetCurrentLine()
            pos = self.GetCurrentPos()
            col = self.GetColumn(pos)
            status = "Line: %d Column: %d Position: %d" % (linenum, col, pos)
            if col == 0:
                status = ""
            self.GetParent().status.show_status_text(status)
            evt.Skip()

    app = wx.App()
    frm = wx.Frame(None,-1,"Test",style=wx.TAB_TRAVERSAL|wx.DEFAULT_FRAME_STYLE,
                   size=(600,400))
    panel = wx.Panel(frm)
    sizer = wx.BoxSizer(wx.HORIZONTAL)

    text = TestSTC(panel, -1)
    text.SetText("Just a placeholder here.\nThe real action is on the borders\nand in the popup status bar.\n\nWhen the text cursor is in any column\nother than zero, the popup status\nbar will show the location.")
    sizer.Add(text, 1, wx.EXPAND)

    panel.status = PopupSubWindow(panel, text)

    def fixFocus(evt):
        evt.Skip()
    text.Bind(wx.EVT_SET_FOCUS, fixFocus)

    panel.SetAutoLayout(True)
    panel.SetSizer(sizer)
    #sizer.Fit(panel)
    #sizer.SetSizeHints(panel)
    panel.Layout()
    app.SetTopWindow(frm)
    frm.Show()
    app.MainLoop()



import collections

import wx

from .action import get_action_id
from . import errors

import logging
log = logging.getLogger("toolbar")
sync_log = logging.getLogger("sync-toolbar")


class ToolbarDescription:
    def __init__(self, parent, editor):
        tb = parent.raw_toolbar
        tb.ClearTools()
        self.valid_id_map = collections.OrderedDict()
        for action_key in editor.toolbar_desc:
            if action_key is None:
                tb.AddSeparator()
            else:
                if action_key.startswith("-"):
                    tb.AddSeparator()
                else:
                    # usable_actions limit the visible actions to what the current editor supports
                    try:
                        action = editor.calc_usable_action(action_key)
                    except KeyError as e:
                        log.debug(f"action {action_key} not used in this editor: {e}")
                        pass
                    else:
                        # a single action can create multiple entries
                        try:
                            action_keys = action.calc_tool_sub_keys(editor)
                            log.debug(f"action {action_key} created subkeys {action_keys}")
                        except AttributeError:
                            action_keys = [action_key]
                        for action_key in action_keys:
                            id = get_action_id(action_key)
                            self.valid_id_map[id] = (action_key, action)
                            action.append_to_toolbar(tb, id, action_key)

    def sync_with_editor(self, toolbar_control):
        for id, (action_key, action) in self.valid_id_map.items():
            sync_log.debug(f"syncing tool {id}: {action_key}, {action}")
            item = toolbar_control.FindById(id)
            try:
                action.sync_tool_item_from_editor(action_key, toolbar_control, id)
            except AttributeError as e:
                sync_log.debug(f"Skipping sync of {action_key} toolbar item from {action.editor}: {e}")
            except errors.RecreateDynamicMenuBar:
                raise
            except:
                log.error(f"Error in sync_with_editor for toolbar item {action_key}")
                raise



import os
import types
import io as BytesIO
import uuid

import numpy as np
import jsonpickle

# Enthought library imports.
from .events import EventHandler
from .utils.command import UndoStack
from .utils import jsonutil
from .utils.nputil import to_numpy
from .utils.pyutil import get_plugins
from .persistence import get_template
from . import filesystem
from .filesystem import fsopen as open
from . import errors

import logging
log = logging.getLogger(__name__)


def find_document_class_for_file(file_metadata):
    """Find the "best" document for a given MIME type string.

    First attempts all documents with exact matches for the MIME string,
    and if no exact matches are found, returns through the list to find
    one that can edit that class of MIME.
    """
    all_documents = get_plugins('sawx.documents', SawxDocument)
    log.debug(f"find_document_class_for_file: file={file_metadata}, known documents: {all_documents}")
    matching_documents = [document for document in all_documents if document.can_load_file_exact(file_metadata)]
    log.debug(f"find_document_class_for_file: exact matches: {matching_documents}")

    if matching_documents:
        return matching_documents[0]

    # Try generic matches if all else fails
    for document in all_documents:
        if document.can_load_file_generic(file_metadata):
            return document
    raise errors.UnsupportedFileType(f"No document available for {file_metadata}")


def identify_document(file_metadata):
    doc_cls = find_document_class_for_file(file_metadata)
    return doc_cls(file_metadata)


class SawxDocument:
    # Class properties

    json_expand_keywords = {}

    session_save_file_extension = ""

    def __init__(self, file_metadata):
        self.undo_stack = UndoStack()
        self.extra_metadata = {}
        self.load(file_metadata)
        self.uuid = str(uuid.uuid4())
        self.change_count = 0
        self.permute = None
        self.baseline_document = None

        # events
        self.refresh_event = EventHandler(self)
        self.recalc_event = EventHandler(self)
        self.structure_changed_event = EventHandler(self)
        self.byte_values_changed_event = EventHandler(self)  # and possibly style, but size of array remains unchanged

        self.byte_style_changed_event = EventHandler(self)  # only styling info may have changed, not any of the data byte values

    def load(self, file_metadata):
        if file_metadata is None:
            self.create_empty()
        else:
            self.file_metadata = file_metadata
            raw = self.load_raw_data()
            self.raw_data = self.calc_raw_data(raw)
            self.load_session()

    def load_raw_data(self):
        fh = open(self.uri, 'rb')
        return fh.read()

    def calc_raw_data(self, raw):
        return to_numpy(raw)

    def create_empty(self):
        self.raw_data = np.zeros(0, dtype=np.uint8)
        self.file_metadata = {'uri': '', 'mime': "application/octet-stream"}

    @property
    def can_revert(self):
        return self.uri != ""

    @property
    def uri(self):
        return self.file_metadata['uri']

    @property
    def mime(self):
        return self.file_metadata['mime']

    @property
    def name(self):
        return os.path.basename(self.uri)

    @property
    def menu_name(self):
        if self.uri:
            return "%s (%s)" % (self.name, self.uri)
        return self.name

    @property
    def root_name(self):
        name, _ = os.path.splitext(self.name)
        return name

    @property
    def extension(self):
        _, ext = os.path.splitext(self.name)
        return ext

    @property
    def is_on_local_filesystem(self):
        try:
            self.filesystem_path()
        except OSError:
            return False
        return True

    @classmethod
    def get_blank(cls):
        return cls(raw_data=b"")

    def __str__(self):
        return f"Document: uuid={self.uuid}, mime={self.mime}, {self.uri}"

    def __len__(self):
        return np.alen(self.raw_data)

    def __getitem__(self, val):
        return self.raw_data[val]

    @property
    def is_dirty(self):
        return self.undo_stack.is_dirty()

    def to_bytes(self):
        return self.raw_data.tostring()

    def load_permute(self, editor):
        if self.permute:
            self.permute.load(self, editor)

    def filesystem_path(self):
        return filesystem.filesystem_path(self.uri)

    @property
    def bytestream(self):
        return BytesIO.BytesIO(self.raw_data)

    # serialization

    def load_session(self):
        d = self.calc_default_session()
        last = self.load_last_session()
        d.update(last)
        self.last_session = d
        self.restore_session(d)

    def calc_default_session(self):
        mime = self.mime
        log.debug(f"calc_default_session: looking for {mime}")
        try:
            text = get_template(mime)
            log.debug(f"calc_default_template: found template for {mime}")
        except OSError:
            log.debug(f"calc_default_template: no template for {mime}")
            e = {}
        else:
            e = jsonutil.unserialize(mime, text)
        return e

    def load_last_session(self):
        session_info = {}
        ext = self.session_save_file_extension
        if ext:
            uri = self.uri + ext
            try:
                fh = open(uri, 'r')
                text = fh.read()
            except IOError:
                log.debug(f"load_last_session: no metadata found at {uri}")
                pass
            else:
                try:
                    session_info = jsonutil.unserialize(uri, text)
                except ValueError as e:
                    log.error(f"invalid data in {uri}: {e}")
        return session_info

    def serialize_session(self, mdict):
        """Save session information to a dict so that it can be serialized
        """
        mdict["document uuid"] = self.uuid
        if self.baseline_document is not None:
            mdict["baseline document"] = self.baseline_document.metadata.uri

    def restore_session(self, e):
        log.debug("restoring sesssion data: %s" % str(e))
        if 'document uuid' in e:
            self.uuid = e['document uuid']
        # if 'baseline document' in e:
        #     try:
        #         self.load_baseline(e['baseline document'])
        #     except DocumentError:
        #         pass
        if 'last_task_id' in e:
            self.last_task_id = e['last_task_id']

    def load_baseline(self, uri, confirm_callback=None):
        log.debug(f"loading baseline data from {uri}")
        if confirm_callback is None:
            confirm_callback = lambda a,b: True
        try:
            guess = FileGuess(uri)
        except Exception as e:
            log.error("Problem loading baseline file %s: %s" % (uri, str(e)))
            raise DocumentError(str(e))
        raw_data = guess.numpy
        difference = len(raw_data) - len(self)
        if difference > 0:
            if confirm_callback("Truncate baseline data by %d bytes?" % difference, "Baseline Size Difference"):
                raw_data = raw_data[0:len(self)]
            else:
                raw_data = []
        elif difference < 0:
            if confirm_callback("Pad baseline data with %d zeros?" % (-difference), "Baseline Size Difference"):
                raw_data = np.pad(raw_data, (0, -difference), "constant", constant_values=0)
            else:
                raw_data = []
        if len(raw_data) > 0:
            self.init_baseline(guess.metadata, raw_data)
        else:
            self.del_baseline()

    def save(self, uri=None):
        if uri is None:
            uri = self.uri
        if not self.verify_writeable_uri(uri):
            raise errors.ReadOnlyFilesystemError(uri)
        if self.verify_ok_to_save():
            raw_data = self.calc_raw_data_to_save()
            self.save_raw_data(uri, raw_data)
            self.file_metadata['uri'] = uri
            self.undo_stack.set_save_point()

    def verify_writeable_uri(self, uri):
        return filesystem.is_user_writeable_uri(uri)

    def verify_ok_to_save(self):
        return True

    def calc_raw_data_to_save(self):
        return self.raw_data.tostring()

    def save_raw_data(self, uri, raw_data):
        fh = open(uri, 'wb')
        log.debug("saving to %s" % uri)
        fh.write(raw_data)
        fh.close()

    def save_session(self, editor_id, editor_session):
        if not self.session_save_file_extension:
            log.debug("no filename extension for session data; not saving")
        else:
            s = {}
            self.serialize_session(s)
            if s and editor_session:
                s[editor_id] = editor_session
                jsonpickle.set_encoder_options("json", sort_keys=True, indent=4)
                text = jsonpickle.dumps(s)
                text = jsonutil.collapse_json(text, 8, self.json_expand_keywords)
                path = self.save_adjacent(self.session_save_file_extension, text)
                log.debug("saved session to {path}")
                return path

    def save_adjacent(self, ext, data, mode="w"):
        if ext:
            path = self.filesystem_path() + ext
            with open(path, mode) as fh:
                fh.write(data)
        else:
            raise RuntimeError(f"Must specify non-blank extension to write a file adjacent to the data file")
        return path

    #### Cleanup functions

    def halt_background_processing(self):
        """Phase 1 of destruction: stop and cleanup any background processing
        before removing any attributes.

        UI elements may by refreshing on timers or in response to background
        processing events, so stop all of that before attempting any other
        attribute cleanup."""
        pass

    def prepare_destroy(self):
        """Phase 2 of destruction: actually destroy resources"""
        pass

    #### file identification

    @classmethod
    def can_load_file_exact(cls, file_metadata):
        return False

    @classmethod
    def can_load_file_generic(cls, file_metadata):
        return False

    @classmethod
    def can_load_file(cls, file_metadata):
        return cls.can_load_file_exact(file_metadata) or cls.can_load_file_generic(file_metadata)



import os

import wx
import wx.html

from ..editor import SawxEditor
from ..filesystem import fsopen as open

import logging
log = logging.getLogger(__name__)


class HtmlWindow(wx.html.HtmlWindow):
    def __init__(self, parent, prefs=None):
        log.debug("calling HtmlWindow constructor")
        wx.html.HtmlWindow.__init__(self, parent, -1, style=wx.NO_FULL_REPAINT_ON_RESIZE)
        log.debug("setting HtmlWindow fonts")
        if prefs:
            self.SetStandardFonts(prefs.font_size, prefs.normal_face, prefs.fixed_face)

    def OnCellMouseHover(self, cell, x, y):
        # Without access to the task window, search the control hierarchy to
        # find a wx.Frame and set the status text directly
        parent = self.GetParent()
        while parent:
            if hasattr(parent, "SetStatusText"):
                linkinfo = cell.GetLink()
                if linkinfo is not None:
                    parent.SetStatusText(linkinfo.GetHref())
                else:
                    parent.SetStatusText("")
                return
            parent = parent.GetParent()


class HtmlViewer(SawxEditor):
    editor_id = "html_viewer"

    ui_name = "HTML Viewer"

    toolbar_desc = [
        "open_file", "save_file", None, "copy"
    ]

    @property
    def can_copy(self):
        return bool(self.control.SelectionToText())

    @property
    def can_paste(self):
        return False

    def create_control(self, parent):
        return HtmlWindow(parent)

    def show(self, args=None):
        self.control.SetPage(self.document.raw_data)

    @classmethod
    def can_edit_document_exact(cls, document):
        return document.mime == "text/html" and document.uri != "about://app"


class TitleScreen(HtmlViewer):
    editor_id = "title_screen"

    ui_name = "Title Screen"

    transient = True

    @classmethod
    def can_edit_document_exact(cls, document):
        return document.mime == "text/html" and document.uri == "about://app"



#!/usr/bin/env python

import sys
import traceback
import wx

import logging
log = logging.getLogger(__name__)


class UndoHistoryPanel(wx.ListBox):
    def __init__(self, parent, editor, **kwargs):
        self.editor = editor
        wx.ListBox.__init__(self, parent, wx.ID_ANY, **kwargs)

        self.Bind(wx.EVT_KEY_DOWN, self.on_key_down)

    def DoGetBestSize(self):
        """ Base class virtual method for sizer use to get the best size
        """
        width = 300
        height = -1
        best = wx.Size(width, height)

        # Cache the best size so it doesn't need to be calculated again,
        # at least until some properties of the window change
        self.CacheBestSize(best)

        return best

    def recalc_view(self):
        self.update_history()

    def refresh_view(self):
        self.Refresh()

    def update_history(self):
        document = self.editor.document
        summary = document.undo_stack.history_list()
        self.Set(summary)
        index = document.undo_stack.insert_index
        if index > 0:
            self.SetSelection(index - 1)

    def activate_spring_tab(self):
        self.recalc_view()

    def get_notification_count(self):
        return 0

    def on_key_down(self, evt):
        key = evt.GetKeyCode()
        log.debug("evt=%s, key=%s" % (evt, key))



import sys

import gc

def find_referrers_of_class(cls):
    gc.collect()
    obj_list = []
    count = 0

    def namestr(obj, namespace):
        return [name for name in namespace if namespace[name] is obj]            
    for obj in gc.get_objects():
        if isinstance(obj, cls):
            print(("Found: %s" % str(obj)))
            rlist = gc.get_referrers(obj)
            print(("  Referenced by: %d" % len(rlist)))
            for r in rlist:
                # print("  * refobj of %s: %s %s" % (obj.__class__.__name__, type(r), namestr(r, globals())))
                print(("  * refobj of %s: %s %s" % (obj.__class__.__name__, type(r), str(r))))
        else:
            count += 1
    print(("%d total objects" % count))



# Standard library imports.
import sys
import weakref

# Major package imports.
import wx

import logging
log = logging.getLogger(__name__)


class MouseMode(object):
    """
    Processing of mouse events, separate from the rendering window
    
    This is an object-based control system of mouse modes
    """
    icon = "help.png"
    menu_item_name = "Generic Mouse Mode"
    menu_item_tooltip = "Tooltip for generic mouse mode"
    editor_trait_for_enabled = ""

    mouse_too_close_pixel_tolerance = 5

    def __init__(self, window):
        self.control = weakref.proxy(window)
        self.snapped_point = None, 0
        self.first_mouse_down_position = 0, 0
        self.after_first_mouse_up = False
        self.mouse_up_too_close = False
        self.can_snap = False

        # left_up events can be called after double click on some platforms, so
        # this parameter can be set by the parent of the handler if the handler
        # wants to differentiate response to left_up
        self.num_clicks = 0

        # Optional (only OS X at this point) mouse wheel event filter
        self.wheel_scroll_count = 0
        self.use_every_nth_wheel_scroll = 5

        self.init_post_hook()

    def init_post_hook(self):
        pass

    def cleanup(self):
        pass

    def get_cursor(self):
        return wx.Cursor(wx.CURSOR_ARROW)

    def get_position(self, evt):
        return evt.GetPosition()

    def reset_early_mouse_params(self):
        self.mouse_up_too_close = False
        self.after_first_mouse_up = False

    def check_early_mouse_release(self, evt):
        c = self.control
        p = evt.GetPosition()
        dx = p[0] - self.first_mouse_down_position[0]
        dy = p[1] - self.first_mouse_down_position[1]
        tol = self.mouse_too_close_pixel_tolerance
        if abs(dx) < tol and abs(dy) < tol:
            return True
        return False

    ##### Sublasses need to override these for customized actions

    def process_mouse_motion_down(self, evt):
        evt.Skip()

    def process_left_down(self, evt):
        evt.Skip()

    def process_mouse_motion_up(self, evt):
        evt.Skip()

    def process_left_up(self, evt):
        evt.Skip()

    def process_left_dclick(self, evt):
        evt.Skip()

    def process_popup(self, evt):
        data = self.calc_popup_data(evt)
        actions = self.calc_popup_actions(evt, data)
        log.debug(f"process_popup: found actions {actions}")
        if actions:
            self.show_popup(actions, data)

    def calc_popup_data(self, evt):
        return None

    def calc_popup_actions(self, evt, data):
        return None

    def show_popup(self, actions):
        raise NotImplementedError("No popup method defined")

    def process_mouse_wheel(self, evt):
        c = self.control
        rotation = evt.GetWheelRotation()
        delta = evt.GetWheelDelta()
        window = evt.GetEventObject()
        log.debug("on_mouse_wheel_scroll. rot=%s delta=%d win=%s" % (rotation, delta, window))
        if rotation == 0 or delta == 0:
            return

        if sys.platform == "darwin":
            # OS X mouse wheel handling is not the same as other platform.
            # The delta value is 10 while the other platforms are 120,
            # and the rotation amount varies, seemingly due to the speed at
            # which the wheel is rotated (or how fast the trackpad is swiped)
            # while other platforms are either 120 or -120.  When mouse wheel
            # handling is performed in the usual manner on OS X it produces a
            # strange back-and-forth zooming in/zooming out.  So, this extra
            # hack is needed to operate like the other platforms.

            # add extra to the rotation so the minimum amount is 1 or -1
            extra = delta if rotation > 0 else -delta
            amount = (rotation + extra) // delta
            self.wheel_scroll_count -= abs(amount)
            if self.wheel_scroll_count > 0:
                return
            self.wheel_scroll_count = self.use_every_nth_wheel_scroll
        else:
            amount = rotation // delta

        if evt.ControlDown():
            if amount < 0:
                self.zoom_out(evt, amount)
            elif amount > 0:
                self.zoom_in(evt, amount)
        else:
            self.pan_mouse_wheel(evt, amount)

    def zoom_out(self, evt, amount):
        evt.Skip()

    def zoom_in(self, evt, amount):
        evt.Skip()

    def pan_mouse_wheel(self, evt, amount):
        self.control.pan_mouse_wheel(evt)

    def process_mouse_enter(self, evt):
        evt.Skip()

    def process_mouse_leave(self, evt):
        evt.Skip()

    def process_focus(self, evt):
        pass

    def process_focus_lost(self, evt):
        pass



""" Simple menubar & tabbed window framework
"""
import time
import argparse
import collections

import wx
import wx.adv

from .frame import SawxFrame, SawxTablessFrame
from .editor import get_editors, find_editor_class_by_id
from .filesystem import init_filesystems
from .filesystem import fsopen as open
from . import persistence
from .events import EventHandler
from .ui import error_logger
from .ui.prefs_dialog import PreferencesDialog
from .utils.background_http import BackgroundHttpDownloader
from . import errors
from .preferences import find_application_preferences

import logging
log = logging.getLogger(__name__)


class SawxApp(wx.App):
    app_name = "Sawx Framework"  # user visible application name

    default_uri = None  # when no files specified on the command line, will default to app_blank_uri

    app_version = "dev"

    app_tagline = "Sawx framework for wxPython applications"

    app_website = "http://playermissile.com/sawx"

    app_author = "Rob McMullen"

    app_icon = "icon://omnivore.ico"

    app_error_email_to = "feedback@playermissile.com"

    app_blank_page = f"""<html>
<h2>{app_name} {app_version}</h2>

<h3>{app_tagline}</h3>
"""

    app_blank_uri = "about://app"

    about_dialog_image = "omnivore256"

    about_dialog_credits = ""

    about_dialog_image_credits = f"""Images & Graphics:<ul>
<li>All the good looking icons are from <a href="https://icons8.com">icons8.com</a>
<li>I needed a few icons too esoteric for a design house, so I did create a few myself... and I'm not an artist. At all.
</ul>
"""

    default_editor = "simple_editor"

    command_line_args = None

    log_dir = ""

    log_file_ext = ".log"

    cache_dir = ""

    user_data_dir = ""

    next_document_id = 0

    documents = []

    clipboard_check_interval = .75

    toolbar_check_interval = .75

    show_progress_bar_on_load = False

    window_sizes = {
        "default": [800, 600],
        "last_window_size": [1000, 800],
    }

    preferences_module = "sawx.preferences"

    _preferences = None

    in_bootup_process = True

    #### Initialization

    def OnInit(self):
        self.SetAppName(self.app_name)
        init_filesystems(self)
        persistence.setup_file_persistence(self.app_name)
        self.remember = persistence.restore_from_last_time()
        self.keybindings_changed_event = EventHandler(self)
        self.init_subprocesses()

        # Initialize dialog-based exception handler
        from .ui import exception_handler

        # Initialize dialog-based progress logger
        from .ui import progress_dialog
        progress_dialog.attach_handler()

        self.active_frame = None
        self.debug_show_focused = False
        self.last_clipboard_check_time = 0
        self.toolbar_timer = wx.Timer(self)
        self.Bind(wx.EVT_TIMER, self.on_timer)
        self.Bind(wx.EVT_IDLE, self.on_idle)
        self.deactivate_app_event = EventHandler(self)
        self.Bind(wx.EVT_ACTIVATE_APP, self.on_activate_app)

        return True

    def OnExit(self):
        persistence.remember_for_next_time(self.remember)
        self.shutdown_subprocesses()
        return wx.App.OnExit(self)

    def process_command_line_args(self, args):
        parser = argparse.ArgumentParser(description="Application parser")
        parser.add_argument("-t", "--task_id", "--task-id", "--edit_with","--edit-with", action="store", default="", help="Use the editing mode specified by this task id for all files listed on the command line")
        parser.add_argument("-d", "--debug_loggers", action="append", nargs=1, help="Comma separated list of debug loggers to enable")
        parser.add_argument("--show_editors", "--show-editors", action="store_true", default=False, help="List all task ids")
        parser.add_argument("--show_focused", "--show-focused", action="store_true", default=False, help="Show the focused window at every idle processing ")
        parser.add_argument("--show_prefs", "--show-prefs", action="store_true", default=False, help="Show the preferences dialog at start time")
        parser.add_argument("--build_docs", "--build-docs", action="store_true", default=False, help="Build documentation from the menubar")
        options, extra_args = parser.parse_known_args(args)
        if options.show_editors:
            for e in get_editors:
                print(f"{e.editor_id}: {e.ui_name}")
        if options.debug_loggers:
            for logger_name in options.debug_loggers:
                error_logger.enable_loggers(logger_name[0])
        task_arguments = collections.OrderedDict()
        if ":" in options.task_id:
            options.task_id, task_str = options.task_id.split(":", 1)
            items = task_str.split(",")
            for item in items:
                if '=' in item:
                    item, v = item.split('=', 1)
                else:
                    v = True
                task_arguments[item] = v
        log.debug("task arguments: %s" % task_arguments)
        try:
            default_editor_cls = find_editor_class_by_id(options.task_id)
        except errors.EditorNotFound:
            default_editor_cls = None
        log.debug(f"default editor: {default_editor_cls}")
        log.debug(f"args: {args}")

        if extra_args:
            log.debug(f"files to load: {extra_args}")
            frame = self.new_frame(uri=self.app_blank_uri, editor_cls=default_editor_cls)
            while len(extra_args) > 0:
                path = extra_args.pop(0)
                frame.load_file(path, None, task_arguments, show_progress_bar=False, default_editor_cls=default_editor_cls)
        else:
            frame = self.new_frame()
        frame.Show()
        if options.show_prefs:
            wx.CallAfter(self.show_preferences_dialog, frame)
        wx.CallAfter(self.done_with_bootup)

    @classmethod
    def done_with_bootup(cls):
        log.debug("done_with_bootup")
        cls.in_bootup_process = False

    def MacOpenFiles(self, filenames):
        """OSX specific routine to handle files that are dropped on the icon
        
        """
        if self.command_line_args is not None:
            # MacOpenFiles gets called for command line arguments, so this flag
            # is used to detect real drops of files onto the dock icon.
            for filename in filenames:
                log.debug(f"MacOpenFiles: loading {filename}")
                self.tasks_application.load_file(filename, None)
        else:
            log.debug(f"MacOpenFiles: skipping {filenames} because it's a command line argument")

    #### Event handlers

    def on_idle(self, evt):
        if self.active_frame is not None:
            t = time.time()
            if t > self.last_clipboard_check_time + self.clipboard_check_interval:
                log.debug("checking clipboard")
                wx.CallAfter(self.on_idle_after)
                self.last_clipboard_check_time = time.time()
                if self.debug_show_focused:
                    self.show_focused()

    def on_idle_after(self):
        try:
            self.active_frame.active_editor.idle_when_active()
            self.active_frame.sync_can_paste()
        except AttributeError as e:
            log.warning(f"on_idle for editor {self.active_frame.active_editor} reported {e}")

    def activate_timer(self, start=True):
        if start:
            log.debug("restarting toolbar timer")
            wx.CallAfter(self.toolbar_timer.Start, self.toolbar_check_interval * 1000)
            log.debug("restarted toolbar timer")
        else:
            log.debug("halting toolbar timer")
            wx.CallAfter(self.toolbar_timer.Stop)
            log.debug("halted toolbar timer")

    def on_activate_app(self, evt):
        log.debug("on_activate_app")
        if evt.GetActive():
            self.activate_timer(True)
        else:
            self.activate_timer(False)
            log.debug("on_activate_app: deactivate!")
            self.deactivate_app_event(evt)

    def on_timer(self, evt):
        evt.Skip()
        if self.active_frame is not None:
            wx.CallAfter(self.active_frame.sync_active_tab)

    #### Shutdown

    def quit(self):
        frames_with_unsaved_stuff = []
        for frame in wx.GetTopLevelWindows():
            try:
                if frame.is_dirty:
                    frames_with_unsaved_stuff.append(frame)
            except AttributeError as e:
                pass
        if frames_with_unsaved_stuff:
            frame = frames_with_unsaved_stuff[0]
            if not frame.confirm("There is unsaved data.\n\nQuit anyway?", "Confirm Loss of Unsaved Data"):
                return

        self.activate_timer(False)
        for frame in wx.GetTopLevelWindows():
            log.debug(f"closing frame {frame}")
            try:
                frame.close_all_tabs()
            except AttributeError as e:
                log.debug(f"error {e}; closing {frame}")
        self.ExitMainLoop()

    #### Application information

    @classmethod
    def get_preferences(cls):
        if cls._preferences is None:
            cls._preferences = find_application_preferences(cls.preferences_module)
            cls._preferences.restore_user_settings()
        return cls._preferences

    @property
    def preferences(self):
        return self.get_preferences()

    @property
    def about_image_bitmap(self):
        data = open(self.about_image, 'rb')
        image = wx.Image(data)
        return wx.Bitmap(image)

    @property
    def about_dialog_html(self):
        return f"""<html>
<body text="#ffffff" bgcolor="#222255" link="#aaaaaa" vlink="#ff0000" alink="#000088">

<h2>{self.app_name} {self.app_version}</h2>

<h3>{self.app_tagline}</h3>

<p>by {self.app_author} <a href="{self.app_website}">{self.app_website}</a>

<p>using {self.about_dialog_system_versions}

<p>{self.about_dialog_credits}

<p>{self.about_dialog_image_credits}
</html>
"""

    @property
    def about_dialog_system_versions(self):
        import sys
        major, minor, micro = sys.version_info[0:3]
        desc = f"Python {major}.{minor}.{micro}:\n<ul>"
        import wx
        desc += "<li>wxPython %s\n" % wx.version()
        try:
            import sawx
            desc += "<li>sawx %s\n" % sawx.__version__
        except:
            pass
        try:
            import numpy
            desc += "<li>numpy %s\n" % numpy.version.version
        except:
            pass
        try:
            import OpenGL
            import OpenGL.GL as gl
            desc += "<li>PyOpenGL %s\n" % OpenGL.__version__
            desc += "<li>OpenGL %s\n" % gl.glGetString(gl.GL_VERSION).encode('utf-8')
            desc += "<li>OpenGL Vendor: %s\n" % gl.glGetString(gl.GL_VENDOR).encode('utf-8')
            desc += "<li>OpenGL Renderer: %s\n" % gl.glGetString(gl.GL_RENDERER).encode('utf-8')
            desc += "<li>GLSL primary: %s\n" % gl.glGetString(gl.GL_SHADING_LANGUAGE_VERSION).encode('utf-8')
            num_glsl = gl.glGetInteger(gl.GL_NUM_SHADING_LANGUAGE_VERSIONS)
            desc += "<li>GLSL supported: "
            for i in range(num_glsl):
                v = gl.glGetStringi(gl.GL_SHADING_LANGUAGE_VERSION, i).encode('utf-8')
                desc += v + ", "
            desc += "\n"
        except:
            pass
        try:
            import gdal
            desc += "<li>GDAL %s\n" % gdal.VersionInfo()
        except:
            pass
        try:
            import pyproj
            desc += "<li>PyProj %s\n" % pyproj.__version__
        except:
            pass
        try:
            import netCDF4
            desc += "<li>netCDF4 %s\n" % netCDF4.__version__
        except:
            pass
        try:
            import shapely
            desc += "<li>Shapely %s\n" % shapely.__version__
        except:
            pass
        try:
            import omnivore_framework
            desc += "<li>Omnivore %s\n" % omnivore_framework.__version__
        except:
            pass
        desc += "</ul>"
        return desc

    def show_about_dialog(self):
        from .ui.dialogs import SawxAboutDialog
        SawxAboutDialog.show_or_raise()

    def show_preferences_dialog(self, parent, page_name=None):
        dialog = PreferencesDialog(parent, page_name)
        dialog.ShowModal()

    #### Convenience functions

    @property
    def last_window_size(self):
        return self.window_sizes["last_window_size"]

    @last_window_size.setter
    def last_window_size(self, value):
        value = list(value)
        log.debug(f"new window size: {value}")
        self.window_sizes["last_window_size"] = value

    def new_frame(self, uri=None, editor_cls=None):
        if uri is None:
            uri = self.default_uri
        if uri is None:
            uri = self.app_blank_uri
        if editor_cls is not None and editor_cls.needs_tabless_frame:
            frame_cls = SawxTablessFrame
        else:
            frame_cls = SawxFrame
        frame = frame_cls(None, uri)
        return frame

    def show_focused(self):
        focused = self.active_frame.FindFocus()
        print(f"Focus at: {focused}" + focused.GetName() if focused is not None else "")

    def find_frame_of_editor(self, editor):
        """Find frame that contains the specified editor"""
        for frame in wx.GetTopLevelWindows():
            try:
                for e in frame.editors:
                    if e == editor:
                        return frame
            except AttributeError:
                pass
        raise KeyError("No frame found containing editor {editor}")

    def find_editor_of_document(self, document):
        """Find editor of the specified document"""
        for frame in wx.GetTopLevelWindows():
            try:
                for e in frame.editors:
                    if e.document == document:
                        return e
            except AttributeError:
                pass
        raise KeyError("No frame found containing editor {editor}")

    #### subprocess helpers

    def init_subprocesses(self):
        self.downloader = None

    def shutdown_subprocesses(self):
        if self.downloader:
            self.downloader.stop_threads()

    def get_downloader(self):
        if self.downloader is None:
            self.downloader = BackgroundHttpDownloader()
        return self.downloader


def restore_from_last_time():
    log.debug("Restoring window sizes")
    cls = wx.GetApp().__class__
    cls.window_sizes = persistence.get_json_data("window_sizes", cls.window_sizes)


def remember_for_next_time():
    log.debug("Remembering window sizes")
    persistence.save_json_data("window_sizes", wx.GetApp().window_sizes)



#!/usr/bin/env python

import wx

import numpy as np


data = [
    (-1, 0, 0, 350),  # root
    (0, 60, 1, 220),  # 1
    (0, 60, 2, 140),
    (0, 60, 3, 180),
    (0, 100, 4, 140),
    (0, 100, 5, 140),  # 5
    (0, 100, 6, 320),
    (0, 200, 7, 240),
    (1, 180, 8, 260),
    (3, 140, 9, 240),
    (0, 200, 10, 260), # 10
    (1, 200, 11, 240),
    (11, 220, 12, 300),
    (11, 220, 13, 320),
    (3, 100, 14, 260),
]


def get_order(source, parent_level):
    """Find restarts with given parent level.

    Building the list from bottom right up, so first restart at the largest
    frame number is first, then later restarts at the same frame number after
    that. Checkpoints that occur at earlier frames then are added, proceeding
    down to frame zero.
    """
    subset_lookup = {}
    for entry in [d for d in source if d[0] == parent_level]:
        start = entry[1]
        if start not in subset_lookup:
            subset_lookup[start] = []
        subset_lookup[start].append(entry)
    processing_order = []
    for items in reversed(list(subset_lookup.values())):
        for item in items:
            processing_order.append(item)
            processing_order.extend(get_order(source, item[2]))
    return processing_order


class RestartLine:
    def __init__(self, data, level):
        self.start_frame = data[1]
        self.end_frame = data[3]
        self.parent = data[0]
        self.restart_number = data[2]
        try:
            current_largest_level = max(level[self.start_frame:self.end_frame + 1])
        except ValueError:
            current_largest_level = 0
        self.level = current_largest_level + 1
        level[self.start_frame:self.end_frame + 1] = self.level
        self.interesting_frames = {
            self.start_frame: "Checkpoint from run {self.parent}",
            self.end_frame: "Last frame",
        }

    def __repr__(self):
        return f"{self.restart_number}: {self.start_frame}->{self.end_frame} @ {self.level}"


class RestartLines:
    def __init__(self, processing_order):
        self.lines = [None] * len(processing_order)
        self.last_frame = max([d[3] for d in processing_order])
        self.highest_level = 0
        self.generate_lines(processing_order)

    def __str__(self):
        txt_lines = []
        for line in self.lines:
            txt_lines.append(repr(line))
        return "\n".join(txt_lines)

    def __iter__(self):
        for line in self.lines:
            yield line

    def __getitem__(self, index):
        return self.lines[index]

    def generate_lines(self, processing_order):
        level = np.zeros(self.last_frame, dtype=np.int16) - 1
        for d in processing_order:
            line = RestartLine(d, level)
            self.lines[line.restart_number] = line
            if line.level > self.highest_level:
                self.highest_level = line.level
            #print(level)
            # print("generate", line, line.level)

    def find_line(self, frame_number, level):
        for line in self.lines:
            if frame_number >= line.start_frame and frame_number <= line.end_frame and level == line.level:
                return line
        return None


wxEVT_CHECKPOINT_RESTORE = wx.NewEventType()
EVT_CHECKPOINT_RESTORE = wx.PyEventBinder(wxEVT_CHECKPOINT_RESTORE, 1)
wxEVT_CHECKPOINT_HOVER = wx.NewEventType()
EVT_CHECKPOINT_HOVER = wx.PyEventBinder(wxEVT_CHECKPOINT_HOVER, 1)

class CheckpointTreeEvent(wx.PyCommandEvent):
    """This event class is used for all events generated by the CheckpointTree
    """
    def __init__(self, type, obj, line, frame_number):
        wx.PyCommandEvent.__init__(self, type)
        self.SetEventObject(obj)
        self.SetId(obj.GetId())
        self.line = line
        self.restart_number = line.restart_number
        self.frame_number = frame_number

    def GetLine(self):
        """The RestartLine of the event
        """
        return self.line

    def GetRestartNumber(self):
        """The restart number of the event
        """
        return self.restart_number

    def GetFrameNumber(self):
        """The frame number of the event
        """
        return self.frame_number


class CheckpointTree(wx.ScrolledWindow):
    def __init__(self, parent, source):
        wx.ScrolledWindow.__init__(self, parent, -1)

        self.width_border = 10
        self.height_border = 10
        self.x_scale = 1
        self.level_height = 10
        self.line_width = 3
        self.marker_size = 3
        self.x_hit = 5
        self.y_hit = 5
        self.over_line = None
        self.scroll_chunk = 20
        self.compute_lines(source)

        self.bg_color = wx.WHITE
        self.line_color = wx.BLUE
        self.marker_color = wx.Colour(255, 255, 255)
        self.cursor_color = wx.Colour(255, 0, 255)
        self.create_dc_stuff()

        self.SetBackgroundColour(self.bg_color)

        self.SetScrollRate(self.scroll_chunk, self.scroll_chunk)

        self.Bind(wx.EVT_PAINT, self.on_paint)
        self.Bind(wx.EVT_SIZE, self.on_size)
        self.Bind(wx.EVT_MOTION, self.on_motion)

    def compute_lines(self, source=None):
        if source is None:
            source = self.source
        self.source = source
        data = source.get_restart_summary()
        # print("DATA", data)
        parent_level = 0
        data.sort()
        order = get_order(data, parent_level)
        order[0:0] = [data[0]]
        # from pprint import pprint
        # pprint(order)
        lines = RestartLines(order)
        print("restart_lines", lines)
        self.set_lines(lines)

    def recalc_view(self):
        self.compute_lines()

    def refresh_view(self):
        self.Refresh()

    def set_lines(self, lines):
        self.restart_lines = lines
        self.max_width = lines.last_frame * self.x_scale + 2 * self.width_border
        self.max_height = self.level_height * lines.highest_level + 2 * self.height_border
        self.virtual_width = self.max_width
        self.virtual_height = self.max_height
        self.SetVirtualSize((self.max_width, self.max_height))

    def create_dc_stuff(self):
        self.pen = wx.Pen(self.line_color, self.line_width)
        self.hover_pen = wx.Pen(self.line_color, self.line_width * 2 + 1)
        self.marker_brush = wx.Brush(self.marker_color)
        self.marker_pen = wx.Pen(self.line_color, 1)
        self.cursor_brush = wx.Brush(self.cursor_color)

    def on_size(self, event):
        size = self.GetVirtualSize()
        self.virtual_width = size.x
        self.virtual_height = size.y

    def on_paint(self, event):
        dc = wx.PaintDC(self)
        x0, y0 = self.CalcUnscrolledPosition (0, 0)
        dc.SetLogicalOrigin(x0, y0)

        dc.SetPen(self.pen)
        for line in self.restart_lines:
            x1 = self.frame_to_x(line.start_frame)
            x2 = self.frame_to_x(line.end_frame)
            y = self.level_to_y(line.level)
            if line == self.over_line:
                dc.SetPen(self.hover_pen)
                dc.DrawLine(x1, y, x2, y)
            else:
                dc.SetPen(self.pen)
                dc.DrawLine(x1, y, x2, y)
            dc.SetPen(self.marker_pen)
            dc.SetBrush(self.marker_brush)
            for interesting_frame in line.interesting_frames.keys():
                xi = self.frame_to_x(interesting_frame)
                dc.DrawCircle(xi, y, self.marker_size)

            if line.restart_number == 0:
                continue

            # draw vertical line connecting to parent restart
            parent_line = self.restart_lines[line.parent]
            parent_y = self.level_to_y(parent_line.level)
            dc.SetPen(self.pen)
            dc.DrawLine(x1, y, x1, parent_y)

        # current frame & restart number
        xc = self.frame_to_x(self.source.current_frame_number)
        restart = self.source.get_restart_of_frame(self.source.current_frame_number)
        line = self.restart_lines[restart.restart_number]
        yc = self.level_to_y(line.level)
        dc.SetPen(self.marker_pen)
        dc.SetBrush(self.cursor_brush)
        dc.DrawCircle(xc, yc, self.marker_size)

    def frame_to_x(self, frame_number):
        return self.width_border + (self.x_scale * frame_number)

    def level_to_y(self, level):
        return self.virtual_height - (self.height_border + level * self.level_height)

    def on_motion(self, evt):
        ex, ey = evt.GetX(), evt.GetY()
        sx, sy = self.GetViewStart()
        x, y = self.CalcUnscrolledPosition (ex, ey)
        size = self.GetVirtualSize()
        # print(ex, sx, x)

        frame_number = (x - self.width_border + self.x_scale // 2) // self.x_scale
        level = ((self.virtual_height - y) - self.height_border + self.level_height // 2) // self.level_height
        x1 = self.frame_to_x(frame_number)
        y1 = self.level_to_y(level)
        # print(x, x1, y, y1, frame_number, level)
        old_over_line = self.over_line
        over_line = None
        if abs(y1 - y) < self.y_hit:
            # print(frame_number, level)
            line = self.restart_lines.find_line(frame_number, level)
            if line:
                print(f"over line {line}")
                over_line = line
                evt = CheckpointTreeEvent(wxEVT_CHECKPOINT_HOVER, self, line, frame_number)
                self.GetEventHandler().ProcessEvent(evt)
        if old_over_line != over_line:
            # print("REFRESHING")
            self.over_line = over_line
            wx.CallAfter(self.Refresh)



#For testing 
if __name__ == '__main__':
    app = wx.App()
    frame = wx.Frame(None, -1, "Test CheckpointTree", size=(500,400))

    from pprint import pprint

    data.sort()
    pprint(data)

    min_frame = min([d[1] for d in data])
    pprint(min_frame)

    max_frame = max([d[3] for d in data])
    pprint(max_frame)

    parent_level = 0
    order = get_order(data, parent_level)
    order[0:0] = [data[0]]
    pprint(order)

    lines = RestartLines(order)
    print("restart_lines", lines)

    class TestSource:
        def get_restart_summary(self):
            return list(data)

    scroll = CheckpointTree(frame, TestSource())
    frame.Show() 
    app.MainLoop()
